@@ -1,6 +1,6 @@
 /*
  * ngx_http_fancyindex_module.c
- * Copyright © 2007-2011 Adrian Perez <aperez@igalia.com>
+ * Copyright © 2007-2013 Adrian Perez <aperez@igalia.com>
  *
  * Module used for fancy indexing of directories. Features and differences
  * with the stock nginx autoindex module:
@@ -40,8 +40,10 @@
  */
 typedef struct {
     ngx_flag_t enable;       /**< Module is enabled. */
+    ngx_uint_t default_sort; /**< Default sort criterion. */
     ngx_flag_t localtime;    /**< File mtime dates are sent in local time. */
     ngx_flag_t exact_size;   /**< Sizes are sent always in bytes. */
+    ngx_uint_t name_length;  /**< Maximum length of file names in bytes. */
 
     ngx_str_t  header;       /**< File name for header, or empty if none. */
     ngx_str_t  footer;       /**< File name for footer, or empty if none. */
@@ -50,9 +52,25 @@ typedef struct {
     ngx_array_t *ignore;     /**< List of files to ignore in listings. */
 } ngx_http_fancyindex_loc_conf_t;
 
+#define NGX_HTTP_FANCYINDEX_SORT_CRITERION_NAME       0
+#define NGX_HTTP_FANCYINDEX_SORT_CRITERION_SIZE       1
+#define NGX_HTTP_FANCYINDEX_SORT_CRITERION_DATE       2
+#define NGX_HTTP_FANCYINDEX_SORT_CRITERION_NAME_DESC  3
+#define NGX_HTTP_FANCYINDEX_SORT_CRITERION_SIZE_DESC  4
+#define NGX_HTTP_FANCYINDEX_SORT_CRITERION_DATE_DESC  5
+
+static ngx_conf_enum_t ngx_http_fancyindex_sort_criteria[] = {
+    { ngx_string("name"), NGX_HTTP_FANCYINDEX_SORT_CRITERION_NAME },
+    { ngx_string("size"), NGX_HTTP_FANCYINDEX_SORT_CRITERION_SIZE },
+    { ngx_string("date"), NGX_HTTP_FANCYINDEX_SORT_CRITERION_DATE },
+    { ngx_string("name_desc"), NGX_HTTP_FANCYINDEX_SORT_CRITERION_NAME_DESC },
+    { ngx_string("size_desc"), NGX_HTTP_FANCYINDEX_SORT_CRITERION_SIZE_DESC },
+    { ngx_string("date_desc"), NGX_HTTP_FANCYINDEX_SORT_CRITERION_DATE_DESC },
+    { ngx_null_string, 0 }
+};
+
 
 #define NGX_HTTP_FANCYINDEX_PREALLOCATE  50
-#define NGX_HTTP_FANCYINDEX_NAME_LEN     50
 
 
 /**
@@ -96,7 +114,17 @@ typedef struct {
 
 
 static int ngx_libc_cdecl
-    ngx_http_fancyindex_cmp_entries(const void *one, const void *two);
+    ngx_http_fancyindex_cmp_entries_name_desc(const void *one, const void *two);
+static int ngx_libc_cdecl
+    ngx_http_fancyindex_cmp_entries_size_desc(const void *one, const void *two);
+static int ngx_libc_cdecl
+    ngx_http_fancyindex_cmp_entries_mtime_desc(const void *one, const void *two);
+static int ngx_libc_cdecl
+    ngx_http_fancyindex_cmp_entries_name_asc(const void *one, const void *two);
+static int ngx_libc_cdecl
+    ngx_http_fancyindex_cmp_entries_size_asc(const void *one, const void *two);
+static int ngx_libc_cdecl
+    ngx_http_fancyindex_cmp_entries_mtime_asc(const void *one, const void *two);
 
 static ngx_int_t ngx_http_fancyindex_error(ngx_http_request_t *r,
     ngx_dir_t *dir, ngx_str_t *name);
@@ -139,6 +167,13 @@ static ngx_command_t  ngx_http_fancyindex_commands[] = {
       offsetof(ngx_http_fancyindex_loc_conf_t, enable),
       NULL },
 
+    { ngx_string("fancyindex_default_sort"),
+      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
+      ngx_conf_set_enum_slot,
+      NGX_HTTP_LOC_CONF_OFFSET,
+      offsetof(ngx_http_fancyindex_loc_conf_t, default_sort),
+      &ngx_http_fancyindex_sort_criteria },
+
     { ngx_string("fancyindex_localtime"),
       NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_FLAG,
       ngx_conf_set_flag_slot,
@@ -153,6 +188,13 @@ static ngx_command_t  ngx_http_fancyindex_commands[] = {
       offsetof(ngx_http_fancyindex_loc_conf_t, exact_size),
       NULL },
 
+    { ngx_string("fancyindex_name_length"),
+      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_FLAG,
+      ngx_conf_set_num_slot,
+      NGX_HTTP_LOC_CONF_OFFSET,
+      offsetof(ngx_http_fancyindex_loc_conf_t, name_length),
+      NULL },
+
     { ngx_string("fancyindex_header"),
       NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_FLAG,
       ngx_conf_set_str_slot,
@@ -364,6 +406,9 @@ make_content_buf(
 {
     ngx_http_fancyindex_entry_t *entry;
 
+    int (*sort_cmp_func) (const void*, const void*);
+    const char  *sort_url_args = "";
+
     off_t        length;
     size_t       len, root, copy, allocated;
     u_char      *filename, *last, scale;
@@ -546,6 +591,7 @@ make_content_buf(
     len = r->uri.len
         + ngx_sizeof_ssz(t05_body2)
         + ngx_sizeof_ssz(t06_list1)
+        + ngx_sizeof_ssz(t_parentdir_entry)
         + ngx_sizeof_ssz(t07_list2)
         ;
 
@@ -555,16 +601,17 @@ make_content_buf(
          * Genearated table rows are as follows, unneeded whitespace
          * is stripped out:
          *
-         *   <tr class="X">
-         *     <td><a href="U">fname</a></td>
+         *   <tr>
+         *     <td><a href="U[?sort]">fname</a></td>
          *     <td>size</td><td>date</td>
          *   </tr>
          */
-        len += ngx_sizeof_ssz("<tr class=\"X\"><td><a href=\"")
+        len += ngx_sizeof_ssz("<tr><td><a href=\"")
             + entry[i].name.len + entry[i].escape /* Escaped URL */
+            + ngx_sizeof_ssz("?C=x&amp;O=y") /* URL sorting arguments */
             + ngx_sizeof_ssz("\">")
             + entry[i].name.len + entry[i].utf_len
-            + NGX_HTTP_FANCYINDEX_NAME_LEN + ngx_sizeof_ssz("&gt;")
+            + alcf->name_length + ngx_sizeof_ssz("&gt;")
             + ngx_sizeof_ssz("</a></td><td>")
             + 20 /* File size */
             + ngx_sizeof_ssz("</td><td>")
@@ -577,11 +624,92 @@ make_content_buf(
     if ((b = ngx_create_temp_buf(r->pool, len)) == NULL)
         return NGX_HTTP_INTERNAL_SERVER_ERROR;
 
+    /*
+     * Determine the sorting criteria. URL arguments look like:
+     *
+     *    C=x[&O=y]
+     *
+     * Where x={M,S,N} and y={A,D}
+     */
+    if ((r->args.len == 3 || (r->args.len == 7 && r->args.data[3] == '&')) &&
+        r->args.data[0] == 'C' && r->args.data[1] == '=')
+    {
+        /* Determine whether the direction of the sorting */
+        ngx_int_t sort_descending = r->args.len == 7
+                                 && r->args.data[4] == 'O'
+                                 && r->args.data[5] == '='
+                                 && r->args.data[6] == 'D';
+
+        /* Pick the sorting criteria */
+        switch (r->args.data[2]) {
+            case 'M': /* Sort by mtime */
+                if (sort_descending) {
+                    sort_cmp_func = ngx_http_fancyindex_cmp_entries_mtime_desc;
+                    if (alcf->default_sort != NGX_HTTP_FANCYINDEX_SORT_CRITERION_DATE_DESC)
+                        sort_url_args = "?C=M&amp;O=D";
+                }
+                else {
+                    sort_cmp_func = ngx_http_fancyindex_cmp_entries_mtime_asc;
+                    if (alcf->default_sort != NGX_HTTP_FANCYINDEX_SORT_CRITERION_DATE)
+                        sort_url_args = "?C=M&amp;O=A";
+                }
+                break;
+            case 'S': /* Sort by size */
+                if (sort_descending) {
+                    sort_cmp_func = ngx_http_fancyindex_cmp_entries_size_desc;
+                    if (alcf->default_sort != NGX_HTTP_FANCYINDEX_SORT_CRITERION_SIZE_DESC)
+                        sort_url_args = "?C=S&amp;O=D";
+                }
+                else {
+                    sort_cmp_func = ngx_http_fancyindex_cmp_entries_size_asc;
+                        if (alcf->default_sort != NGX_HTTP_FANCYINDEX_SORT_CRITERION_SIZE)
+                    sort_url_args = "?C=S&amp;O=A";
+                }
+                break;
+            case 'N': /* Sort by name */
+            default:
+                if (sort_descending) {
+                    sort_cmp_func = ngx_http_fancyindex_cmp_entries_name_desc;
+                    if (alcf->default_sort != NGX_HTTP_FANCYINDEX_SORT_CRITERION_NAME_DESC)
+                        sort_url_args = "?C=N&amp;O=D";
+                }
+                else {
+                    sort_cmp_func = ngx_http_fancyindex_cmp_entries_name_asc;
+                    if (alcf->default_sort != NGX_HTTP_FANCYINDEX_SORT_CRITERION_NAME)
+                        sort_url_args = "?C=N&amp;O=A";
+                }
+                break;
+        }
+    }
+    else {
+        switch (alcf->default_sort) {
+            case NGX_HTTP_FANCYINDEX_SORT_CRITERION_DATE_DESC:
+                sort_cmp_func = ngx_http_fancyindex_cmp_entries_mtime_desc;
+                break;
+            case NGX_HTTP_FANCYINDEX_SORT_CRITERION_DATE:
+                sort_cmp_func = ngx_http_fancyindex_cmp_entries_mtime_asc;
+                break;
+            case NGX_HTTP_FANCYINDEX_SORT_CRITERION_SIZE_DESC:
+                sort_cmp_func = ngx_http_fancyindex_cmp_entries_size_desc;
+                break;
+            case NGX_HTTP_FANCYINDEX_SORT_CRITERION_SIZE:
+                sort_cmp_func = ngx_http_fancyindex_cmp_entries_size_asc;
+                break;
+            case NGX_HTTP_FANCYINDEX_SORT_CRITERION_NAME_DESC:
+                sort_cmp_func = ngx_http_fancyindex_cmp_entries_name_desc;
+                break;
+            case NGX_HTTP_FANCYINDEX_SORT_CRITERION_NAME:
+            default:
+                sort_cmp_func = ngx_http_fancyindex_cmp_entries_name_asc;
+                break;
+        }
+    }
+
     /* Sort entries, if needed */
     if (entries.nelts > 1) {
         ngx_qsort(entry, (size_t) entries.nelts,
-                sizeof(ngx_http_fancyindex_entry_t),
-                ngx_http_fancyindex_cmp_entries);
+                  sizeof(ngx_http_fancyindex_entry_t),
+                  sort_cmp_func);
     }
 
     b->last = ngx_cpymem_str(b->last, r->uri);
@@ -590,15 +718,24 @@ make_content_buf(
 
     tp = ngx_timeofday();
 
+    /* "Parent dir" entry, always first */
+    b->last = ngx_cpymem_ssz(b->last,
+                             "<tr>"
+                             "<td><a href=\"../");
+    if (*sort_url_args) {
+        b->last = ngx_cpymem(b->last,
+                             sort_url_args,
+                             ngx_sizeof_ssz("?C=N&amp;O=A"));
+    }
+    b->last = ngx_cpymem_ssz(b->last,
+                             "\">Parent directory/</a></td>"
+                             "<td>-</td>"
+                             "<td>-</td>"
+                             "</tr>");
+
+    /* Entries for directories and files */
     for (i = 0; i < entries.nelts; i++) {
-        static const char _evenodd[] = { 'e', 'o' };
-        b->last = ngx_cpymem_ssz(b->last, "<tr class=\"");
-        *b->last++ = _evenodd[i & 0x01];
-        /*
-         * Alternative implementation:
-         *   *b->last++ = (i & 0x01) ? 'e' : 'o';
-         */
-        b->last = ngx_cpymem_ssz(b->last, "\"><td><a href=\"");
+        b->last = ngx_cpymem_ssz(b->last, "<tr><td><a href=\"");
 
         if (entry[i].escape) {
             ngx_fancyindex_escape_uri(b->last,
@@ -613,6 +750,11 @@ make_content_buf(
 
         if (entry[i].dir) {
             *b->last++ = '/';
+            if (*sort_url_args) {
+                b->last = ngx_cpymem(b->last,
+                                     sort_url_args,
+                                     ngx_sizeof_ssz("?C=x&amp;O=y"));
+            }
         }
 
         *b->last++ = '"';
@@ -621,10 +763,10 @@ make_content_buf(
         len = entry[i].utf_len;
 
         if (entry[i].name.len - len) {
-            if (len > NGX_HTTP_FANCYINDEX_NAME_LEN) {
-                copy = NGX_HTTP_FANCYINDEX_NAME_LEN - 3 + 1;
+            if (len > alcf->name_length) {
+                copy = alcf->name_length - 3 + 1;
             } else {
-                copy = NGX_HTTP_FANCYINDEX_NAME_LEN + 1;
+                copy = alcf->name_length + 1;
             }
 
             b->last = ngx_utf8_cpystrn(b->last, entry[i].name.data,
@@ -633,15 +775,15 @@ make_content_buf(
 
         } else {
             b->last = ngx_cpystrn(b->last, entry[i].name.data,
-                                  NGX_HTTP_FANCYINDEX_NAME_LEN + 1);
+                                  alcf->name_length + 1);
             last = b->last - 3;
         }
 
-        if (len > NGX_HTTP_FANCYINDEX_NAME_LEN) {
+        if (len > alcf->name_length) {
             b->last = ngx_cpymem_ssz(last, "..&gt;</a></td><td>");
 
         } else {
-            if (entry[i].dir && NGX_HTTP_FANCYINDEX_NAME_LEN - len > 0) {
+            if (entry[i].dir && alcf->name_length - len > 0) {
                 *b->last++ = '/';
                 len++;
             }
@@ -898,18 +1040,70 @@ add_builtin_header:
 
 
 static int ngx_libc_cdecl
-ngx_http_fancyindex_cmp_entries(const void *one, const void *two)
+ngx_http_fancyindex_cmp_entries_name_desc(const void *one, const void *two)
 {
     ngx_http_fancyindex_entry_t *first = (ngx_http_fancyindex_entry_t *) one;
     ngx_http_fancyindex_entry_t *second = (ngx_http_fancyindex_entry_t *) two;
 
+    /* move the directories to the start */
+    if (first->dir && !second->dir) {
+        return -1;
+    }
+    if (!first->dir && second->dir) {
+        return 1;
+    }
+
+    return (int) ngx_strcmp(second->name.data, first->name.data);
+}
+
+
+static int ngx_libc_cdecl
+ngx_http_fancyindex_cmp_entries_size_desc(const void *one, const void *two)
+{
+    ngx_http_fancyindex_entry_t *first = (ngx_http_fancyindex_entry_t *) one;
+    ngx_http_fancyindex_entry_t *second = (ngx_http_fancyindex_entry_t *) two;
+
+    /* move the directories to the start */
+    if (first->dir && !second->dir) {
+        return -1;
+    }
+    if (!first->dir && second->dir) {
+        return 1;
+    }
+
+    return second->size - first->size;
+}
+
+
+static int ngx_libc_cdecl
+ngx_http_fancyindex_cmp_entries_mtime_desc(const void *one, const void *two)
+{
+    ngx_http_fancyindex_entry_t *first = (ngx_http_fancyindex_entry_t *) one;
+    ngx_http_fancyindex_entry_t *second = (ngx_http_fancyindex_entry_t *) two;
+
+    /* move the directories to the start */
+    if (first->dir && !second->dir) {
+        return -1;
+    }
+    if (!first->dir && second->dir) {
+        return 1;
+    }
+
+    return second->mtime - first->mtime;
+}
+
+
+static int ngx_libc_cdecl
+ngx_http_fancyindex_cmp_entries_name_asc(const void *one, const void *two)
+{
+    ngx_http_fancyindex_entry_t *first = (ngx_http_fancyindex_entry_t *) one;
+    ngx_http_fancyindex_entry_t *second = (ngx_http_fancyindex_entry_t *) two;
+
+    /* move the directories to the start */
     if (first->dir && !second->dir) {
-        /* move the directories to the start */
         return -1;
     }
-
     if (!first->dir && second->dir) {
-        /* move the directories to the start */
         return 1;
     }
 
@@ -917,6 +1111,40 @@ ngx_http_fancyindex_cmp_entries(const void *one, const void *two)
 }
 
 
+static int ngx_libc_cdecl
+ngx_http_fancyindex_cmp_entries_size_asc(const void *one, const void *two)
+{
+    ngx_http_fancyindex_entry_t *first = (ngx_http_fancyindex_entry_t *) one;
+    ngx_http_fancyindex_entry_t *second = (ngx_http_fancyindex_entry_t *) two;
+
+    /* move the directories to the start */
+    if (first->dir && !second->dir) {
+        return -1;
+    }
+    if (!first->dir && second->dir) {
+        return 1;
+    }
+
+    return first->size - second->size;
+}
+
+
+static int ngx_libc_cdecl
+ngx_http_fancyindex_cmp_entries_mtime_asc(const void *one, const void *two)
+{
+    ngx_http_fancyindex_entry_t *first = (ngx_http_fancyindex_entry_t *) one;
+    ngx_http_fancyindex_entry_t *second = (ngx_http_fancyindex_entry_t *) two;
+
+    /* move the directories to the start */
+    if (first->dir && !second->dir) {
+        return -1;
+    }
+    if (!first->dir && second->dir) {
+        return 1;
+    }
+
+    return first->mtime - second->mtime;
+}
 
 
 static ngx_int_t
@@ -948,10 +1176,12 @@ ngx_http_fancyindex_create_loc_conf(ngx_conf_t *cf)
      *    conf->footer.len   = 0
      *    conf->footer.data  = NULL
      */
-    conf->enable = NGX_CONF_UNSET;
-    conf->localtime = NGX_CONF_UNSET;
-    conf->exact_size = NGX_CONF_UNSET;
-    conf->ignore = NGX_CONF_UNSET_PTR;
+    conf->enable       = NGX_CONF_UNSET;
+    conf->default_sort = NGX_CONF_UNSET_UINT;
+    conf->localtime    = NGX_CONF_UNSET;
+    conf->name_length  = NGX_CONF_UNSET_UINT;
+    conf->exact_size   = NGX_CONF_UNSET;
+    conf->ignore       = NGX_CONF_UNSET_PTR;
 
     return conf;
 }
@@ -964,8 +1194,10 @@ ngx_http_fancyindex_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
     ngx_http_fancyindex_loc_conf_t *conf = child;
 
     ngx_conf_merge_value(conf->enable, prev->enable, 0);
+    ngx_conf_merge_uint_value(conf->default_sort, prev->default_sort, NGX_HTTP_FANCYINDEX_SORT_CRITERION_NAME);
     ngx_conf_merge_value(conf->localtime, prev->localtime, 0);
     ngx_conf_merge_value(conf->exact_size, prev->exact_size, 1);
+    ngx_conf_merge_uint_value(conf->name_length, prev->name_length, 50);
 
     ngx_conf_merge_str_value(conf->header, prev->header, "");
     ngx_conf_merge_str_value(conf->footer, prev->footer, "");
